# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from neutron_lib.api.definitions import constants as fwaas_consts
from neutron_lib.plugins import directory
from oslo_utils import uuidutils

from vmware_nsx.services.fwaas.nsx_p import edge_fwaas_driver_v2
from vmware_nsx.services.fwaas.nsx_p import fwaas_callbacks_v2
from vmware_nsx.tests.unit.nsx_p import test_plugin as test_p_plugin
from vmware_nsxlib.v3 import nsx_constants as consts
from vmware_nsxlib.v3.policy import constants as policy_constants

FAKE_FW_ID = 'fake_fw_uuid'
FAKE_ROUTER_ID = 'fake_rtr_uuid'
FAKE_PORT_ID = 'fake_port_uuid'
FAKE_NET_ID = 'fake_net_uuid'
GW_POLICY_PATH = ("vmware_nsxlib.v3.policy.core_resources."
                  "NsxPolicyGatewayPolicyApi")


class NsxpFwaasTestCase(test_p_plugin.NsxPPluginTestCaseMixin):
    def setUp(self):
        super(NsxpFwaasTestCase, self).setUp()
        self.firewall = edge_fwaas_driver_v2.EdgeFwaasPDriverV2()

        self.project_id = uuidutils.generate_uuid()
        self.plugin = directory.get_plugin()
        self.plugin.fwaas_callbacks = fwaas_callbacks_v2.NsxpFwaasCallbacksV2(
            False)
        self.plugin.fwaas_callbacks.fwaas_enabled = True
        self.plugin.fwaas_callbacks.fwaas_driver = self.firewall
        self.plugin.fwaas_callbacks.internal_driver = self.firewall
        self.plugin.init_is_complete = True

        def mock_get_random_rule_id(rid):
            return rid

        mock.patch.object(self.plugin.fwaas_callbacks, '_get_random_rule_id',
                          side_effect=mock_get_random_rule_id).start()

        mock.patch.object(self.plugin.nsxpolicy, 'search_by_tags',
                          return_value={'results': []}).start()

    def _default_rule(self, seq_num):
        return self.plugin.nsxpolicy.gateway_policy.build_entry(
                fwaas_callbacks_v2.DEFAULT_RULE_NAME,
                policy_constants.DEFAULT_DOMAIN, FAKE_ROUTER_ID,
                fwaas_callbacks_v2.DEFAULT_RULE_ID,
                description=fwaas_callbacks_v2.DEFAULT_RULE_NAME,
                action=consts.FW_ACTION_ALLOW,
                scope=[self.plugin.nsxpolicy.tier1.get_path(FAKE_ROUTER_ID)],
                sequence_number=seq_num,
                direction=consts.IN_OUT).get_obj_dict()

    def _block_interface_rules(self, seq_num):
        net_group_id = '%s-%s' % (FAKE_ROUTER_ID, FAKE_NET_ID)
        ingress_rule = self.plugin.nsxpolicy.gateway_policy.build_entry(
                "Block port ingress",
                policy_constants.DEFAULT_DOMAIN, FAKE_ROUTER_ID,
                fwaas_callbacks_v2.DEFAULT_RULE_ID + FAKE_NET_ID + 'ingress',
                action=consts.FW_ACTION_DROP,
                dest_groups=[net_group_id],
                scope=[self.plugin.nsxpolicy.tier1.get_path(FAKE_ROUTER_ID)],
                sequence_number=seq_num,
                direction=consts.IN)

        egress_rule = self.plugin.nsxpolicy.gateway_policy.build_entry(
                "Block port egress",
                policy_constants.DEFAULT_DOMAIN, FAKE_ROUTER_ID,
                fwaas_callbacks_v2.DEFAULT_RULE_ID + FAKE_NET_ID + 'egress',
                action=consts.FW_ACTION_DROP,
                source_groups=[net_group_id],
                scope=[self.plugin.nsxpolicy.tier1.get_path(FAKE_ROUTER_ID)],
                sequence_number=seq_num + 1,
                direction=consts.OUT)

        return [ingress_rule.get_obj_dict(), egress_rule.get_obj_dict()]

    def _fake_rules_v4(self, is_ingress=True, cidr='10.24.4.0/24',
                       is_conflict=False):
        rule1 = {'enabled': True,
                 'action': 'allow',
                 'ip_version': 4,
                 'protocol': 'tcp',
                 'destination_port': '80',
                 'id': 'fake-fw-rule1',
                 'description': 'first rule'}
        rule2 = {'name': 'rule 2',
                 'enabled': True,
                 'action': 'reject',
                 'ip_version': 4,
                 'protocol': 'tcp',
                 'destination_port': '22:24',
                 'source_port': '1:65535',
                 'id': 'fake-fw-rule2'}
        rule3 = {'enabled': True,
                 'action': 'deny',
                 'ip_version': 4,
                 'protocol': 'icmp',
                 'id': 'fake-fw-rule3'}
        rule4 = {'enabled': True,
                 'action': 'deny',
                 'ip_version': 4,
                 'id': 'fake-fw-rule4'}
        if is_ingress:
            if not is_conflict:
                rule1['source_ip_address'] = cidr
            else:
                rule1['destination_ip_address'] = cidr
        else:
            if not is_conflict:
                rule1['destination_ip_address'] = cidr
            else:
                rule1['source_ip_address'] = cidr

        return [rule1, rule2, rule3, rule4]

    def _translated_cidr(self, cidr):
        if cidr is None:
            return []
        else:
            return [{'target_id': cidr,
                     'target_type': 'IPv4Address'}]

    def _validate_rules_translation(self, actual_rules, rule_list, is_ingress):
        for index in range(len(rule_list)):
            self._validate_rule_translation(
                actual_rules[index].get_obj_dict(),
                rule_list[index],
                is_ingress)

    def _validate_rule_translation(self, nsx_rule, fw_rule, is_ingress):
        self.assertEqual(fw_rule['id'], nsx_rule['id'])
        self.assertEqual(fwaas_callbacks_v2.RULE_NAME_PREFIX +
                         (fw_rule.get('name') or fw_rule['id']),
                         nsx_rule['display_name'])
        self.assertEqual(fw_rule.get('description'), nsx_rule['description'])
        self.assertEqual(consts.IN if is_ingress else consts.OUT,
                         nsx_rule['direction'])
        self.assertEqual(self.plugin.nsxpolicy.tier1.get_path(FAKE_ROUTER_ID),
                         nsx_rule['scope'][0])

        # Action
        if (fw_rule['action'] == fwaas_consts.FWAAS_REJECT or
            fw_rule['action'] == fwaas_consts.FWAAS_DENY):
            self.assertEqual(consts.FW_ACTION_DROP, nsx_rule['action'])
        else:
            self.assertEqual(consts.FW_ACTION_ALLOW, nsx_rule['action'])

        # Service
        if fw_rule.get('protocol') in ['tcp', 'udp', 'icmp']:
            self.assertEqual(1, len(nsx_rule['services']))
            self.assertTrue(
                nsx_rule['services'][0].startswith('/infra/services/'))
        # Source & destination
        if (fw_rule.get('source_ip_address') and
            not fw_rule['source_ip_address'].startswith('0.0.0.0')):
            self.assertEqual(['/infra/domains/%s/groups/source-%s' % (
                                policy_constants.DEFAULT_DOMAIN,
                                fw_rule['id'])],
                             nsx_rule['source_groups'])
        elif not is_ingress:
            self.assertEqual(['/infra/domains/%s/groups/%s-%s' % (
                                policy_constants.DEFAULT_DOMAIN,
                                FAKE_ROUTER_ID, FAKE_NET_ID)],
                             nsx_rule['source_groups'])

        if (fw_rule.get('destination_ip_address') and
            not fw_rule['destination_ip_address'].startswith('0.0.0.0')):
            self.assertEqual(['/infra/domains/%s/groups/destination-%s' % (
                                policy_constants.DEFAULT_DOMAIN,
                                fw_rule['id'])],
                             nsx_rule['destination_groups'])
        elif is_ingress:
            self.assertEqual(['/infra/domains/%s/groups/%s-%s' % (
                                policy_constants.DEFAULT_DOMAIN,
                                FAKE_ROUTER_ID, FAKE_NET_ID)],
                             nsx_rule['destination_groups'])

    def _fake_empty_firewall_group(self):
        fw_inst = {'id': FAKE_FW_ID,
                   'admin_state_up': True,
                   'tenant_id': self.project_id,
                   'ingress_rule_list': [],
                   'egress_rule_list': []}
        return fw_inst

    def _fake_firewall_group(self, rule_list, is_ingress=True,
                             admin_state_up=True):
        _rule_list = copy.deepcopy(rule_list)
        for rule in _rule_list:
            rule['position'] = str(_rule_list.index(rule))
        fw_inst = {'id': FAKE_FW_ID,
                   'admin_state_up': admin_state_up,
                   'tenant_id': self.project_id,
                   'ingress_rule_list': [],
                   'egress_rule_list': []}
        if is_ingress:
            fw_inst['ingress_rule_list'] = _rule_list
        else:
            fw_inst['egress_rule_list'] = _rule_list
        return fw_inst

    def _fake_firewall_group_with_admin_down(self, rule_list,
                                             is_ingress=True):
        return self._fake_firewall_group(
            rule_list, is_ingress=is_ingress, admin_state_up=False)

    def _fake_apply_list(self):
        router_inst = {'id': FAKE_ROUTER_ID, 'external_gateway_info': 'dummy'}
        router_info_inst = mock.Mock()
        router_info_inst.router = router_inst
        router_info_inst.router_id = FAKE_ROUTER_ID
        apply_list = [(router_info_inst, FAKE_PORT_ID)]
        return apply_list

    def test_create_firewall_no_rules(self):
        apply_list = self._fake_apply_list()
        firewall = self._fake_empty_firewall_group()
        port = {'id': FAKE_PORT_ID, 'network_id': FAKE_NET_ID}
        with mock.patch.object(self.plugin, '_get_router_interfaces',
                               return_value=[port]),\
            mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=firewall),\
            mock.patch.object(self.plugin, '_get_router',
                              return_value={'project_id': self.project_id}),\
            mock.patch.object(self.plugin, 'service_router_has_services',
                              return_value=True),\
            mock.patch(GW_POLICY_PATH + ".update_entries") as update_fw:
            self.firewall.create_firewall_group('nsx', apply_list, firewall)
            # expecting 2 block rules for the logical switch (egress & ingress)
            # and last default allow all rule
            expected_rules = (self._block_interface_rules(0) +
                              [self._default_rule(2)])
            update_fw.assert_called_once_with(
                policy_constants.DEFAULT_DOMAIN, FAKE_ROUTER_ID, mock.ANY,
                category=policy_constants.CATEGORY_LOCAL_GW)
            # compare rules one by one
            actual_rules = update_fw.call_args[0][2]
            self.assertEqual(len(expected_rules), len(actual_rules))
            for index in range(len(actual_rules)):
                self.assertEqual(expected_rules[index],
                                 actual_rules[index].get_obj_dict())

    def _setup_firewall_with_rules(self, func, is_ingress=True,
                                   is_conflict=False, cidr='10.24.4.0/24'):
        apply_list = self._fake_apply_list()
        rule_list = self._fake_rules_v4(is_ingress=is_ingress,
                                        is_conflict=is_conflict,
                                        cidr=cidr)
        firewall = self._fake_firewall_group(rule_list, is_ingress=is_ingress)
        port = {'id': FAKE_PORT_ID, 'network_id': FAKE_NET_ID}
        with mock.patch.object(self.plugin, '_get_router_interfaces',
                               return_value=[port]),\
            mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=firewall), \
            mock.patch.object(self.plugin, '_get_router',
                              return_value={'project_id': self.project_id}),\
            mock.patch.object(self.plugin, 'service_router_has_services',
                              return_value=True), \
            mock.patch(GW_POLICY_PATH + ".update_entries") as update_fw:
            func('nsx', apply_list, firewall)
            expected_default_rules = self._block_interface_rules(
                len(rule_list)) + [self._default_rule(len(rule_list) + 2)]
            update_fw.assert_called_once_with(
                policy_constants.DEFAULT_DOMAIN, FAKE_ROUTER_ID, mock.ANY,
                category=policy_constants.CATEGORY_LOCAL_GW)

            # compare rules one by one
            actual_rules = update_fw.call_args[0][2]
            self.assertEqual(len(rule_list) + 3, len(actual_rules))
            self._validate_rules_translation(
                actual_rules,
                rule_list,
                is_ingress)
            # compare the last 3 rules (default interface rules +
            # default allow rule)
            self.assertEqual(actual_rules[-3].get_obj_dict(),
                             expected_default_rules[0])
            self.assertEqual(actual_rules[-2].get_obj_dict(),
                             expected_default_rules[1])
            self.assertEqual(actual_rules[-1].get_obj_dict(),
                             expected_default_rules[2])

    def test_create_firewall_with_ingress_rules(self):
        self._setup_firewall_with_rules(self.firewall.create_firewall_group)

    def test_update_firewall_with_ingress_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group)

    def test_create_firewall_with_egress_rules(self):
        self._setup_firewall_with_rules(self.firewall.create_firewall_group,
                                        is_ingress=False)

    def test_update_firewall_with_egress_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group,
                                        is_ingress=False)

    def test_create_firewall_with_egress_conflicting_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group,
                                        is_ingress=False, is_conflict=True)

    def test_create_firewall_with_ingress_conflicting_rules(self):
        self._setup_firewall_with_rules(self.firewall.update_firewall_group,
                                        is_ingress=True, is_conflict=True)

    def test_create_firewall_with_illegal_cidr(self):
        self._setup_firewall_with_rules(self.firewall.create_firewall_group,
                                        cidr='0.0.0.0/24')

    def test_delete_firewall(self):
        apply_list = self._fake_apply_list()
        firewall = self._fake_empty_firewall_group()
        port = {'id': FAKE_PORT_ID}
        with mock.patch.object(self.plugin, '_get_router_interfaces',
                               return_value=[port]),\
            mock.patch.object(self.plugin.fwaas_callbacks, 'get_port_fwg',
                              return_value=None), \
            mock.patch.object(self.plugin, '_get_router',
                              return_value={'project_id': self.project_id}),\
            mock.patch.object(self.plugin, 'service_router_has_services',
                              return_value=True), \
            mock.patch(GW_POLICY_PATH + ".delete") as delete_fw:
            self.firewall.delete_firewall_group('nsx', apply_list, firewall)
            delete_fw.assert_called_once_with(
                policy_constants.DEFAULT_DOMAIN, map_id=FAKE_ROUTER_ID)

    def test_create_firewall_with_admin_down(self):
        apply_list = self._fake_apply_list()
        rule_list = self._fake_rules_v4()
        firewall = self._fake_firewall_group_with_admin_down(rule_list)
        with mock.patch.object(self.plugin, 'service_router_has_services',
                               return_value=True), \
            mock.patch.object(self.plugin, '_get_router',
                              return_value={'project_id': self.project_id}),\
            mock.patch(GW_POLICY_PATH + ".create_with_entries") as update_fw:
            self.firewall.create_firewall_group('nsx', apply_list, firewall)
            update_fw.assert_not_called()



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging

from vmware_nsx._i18n import _
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_p.implementation import lb_const as p_const
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsxlib.v3 import load_balancer as nsxlib_lb
from vmware_nsxlib.v3.policy import constants as p_constants
from vmware_nsxlib.v3.policy import utils as p_utils
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)
ADV_RULE_NAME = 'LB external VIP advertisement'


def get_rule_match_conditions(policy):
    match_conditions = []
    # values in rule have already been validated in LBaaS API,
    # we won't need to valid anymore in driver, and just get
    # the LB rule mapping from the dict.
    for rule in policy['rules']:
        match_type = lb_const.LB_RULE_MATCH_TYPE[rule['compare_type']]
        if rule['type'] == lb_const.L7_RULE_TYPE_COOKIE:
            header_value = rule['key'] + '=' + rule['value']
            match_conditions.append(
                {'type': 'LBHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': 'Cookie',
                 'header_value': header_value})
        elif rule['type'] == lb_const.L7_RULE_TYPE_FILE_TYPE:
            match_conditions.append(
                {'type': 'LBHttpRequestUriCondition',
                 'match_type': match_type,
                 'uri': '*.' + rule['value']})
        elif rule['type'] == lb_const.L7_RULE_TYPE_HEADER:
            match_conditions.append(
                {'type': 'LBHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': rule['key'],
                 'header_value': rule['value']})
        elif rule['type'] == lb_const.L7_RULE_TYPE_HOST_NAME:
            match_conditions.append(
                {'type': 'LBHttpRequestHeaderCondition',
                 'match_type': match_type,
                 'header_name': 'Host',
                 'header_value': rule['value']})
        elif rule['type'] == lb_const.L7_RULE_TYPE_PATH:
            match_conditions.append(
                {'type': 'LBHttpRequestUriCondition',
                 'match_type': match_type,
                 'uri': rule['value']})
        else:
            msg = (_('l7rule type %(type)s is not supported in LBaaS') %
                   {'type': rule['type']})
            raise n_exc.BadRequest(resource='lbaas-l7rule', msg=msg)
    return match_conditions


def get_rule_actions(nsxpolicy, l7policy):
    if l7policy['action'] == lb_const.L7_POLICY_ACTION_REDIRECT_TO_POOL:
        if l7policy['redirect_pool_id']:
            lb_pool_id = l7policy['redirect_pool_id']
            lb_pool_path = nsxpolicy.load_balancer.lb_pool.get_path(lb_pool_id)
            actions = [{'type': p_const.LB_SELECT_POOL_ACTION,
                        'pool_id': lb_pool_path}]
        else:
            msg = _('Failed to get LB pool binding from nsx db')
            raise n_exc.BadRequest(resource='lbaas-l7rule-create',
                                   msg=msg)
    elif l7policy['action'] == lb_const.L7_POLICY_ACTION_REDIRECT_TO_URL:
        actions = [{'type': p_const.LB_HTTP_REDIRECT_ACTION,
                    'redirect_status': lb_const.LB_HTTP_REDIRECT_STATUS,
                    'redirect_url': l7policy['redirect_url']}]
    elif l7policy['action'] == lb_const.L7_POLICY_ACTION_REJECT:
        actions = [{'type': p_const.LB_REJECT_ACTION,
                    'reply_status': lb_const.LB_HTTP_REJECT_STATUS}]
    else:
        msg = (_('Invalid l7policy action: %(action)s') %
               {'action': l7policy['action']})
        raise n_exc.BadRequest(resource='lbaas-l7rule-create',
                               msg=msg)
    return actions


def convert_l7policy_to_lb_rule(nsxpolicy, policy):
    return {
        'match_conditions': get_rule_match_conditions(policy),
        'actions': get_rule_actions(nsxpolicy, policy),
        'phase': lb_const.LB_RULE_HTTP_FORWARDING,
        'match_strategy': 'ALL'
    }


def remove_rule_from_policy(rule):
    l7rules = rule['policy']['rules']
    rule['policy']['rules'] = [r for r in l7rules if r['id'] != rule['id']]


def update_rule_in_policy(rule):
    remove_rule_from_policy(rule)
    rule['policy']['rules'].append(rule)


def update_router_lb_vip_advertisement(context, core_plugin, router_id):
    router = core_plugin.get_router(context.elevated(), router_id)

    # Add a rule to advertise external vips on the router
    external_subnets = core_plugin._find_router_gw_subnets(
        context.elevated(), router)
    external_cidrs = [s['cidr'] for s in external_subnets]
    if external_cidrs:
        core_plugin.nsxpolicy.tier1.add_advertisement_rule(
            router_id,
            ADV_RULE_NAME,
            p_constants.ADV_RULE_PERMIT,
            p_constants.ADV_RULE_OPERATOR_GE,
            [p_constants.ADV_RULE_TIER1_LB_VIP],
            external_cidrs)


def get_monitor_policy_client(lb_client, hm):
    if hm['type'] == lb_const.LB_HEALTH_MONITOR_TCP:
        return lb_client.lb_monitor_profile_tcp
    elif hm['type'] == lb_const.LB_HEALTH_MONITOR_HTTP:
        return lb_client.lb_monitor_profile_http
    elif hm['type'] == lb_const.LB_HEALTH_MONITOR_HTTPS:
        return lb_client.lb_monitor_profile_https
    elif hm['type'] == lb_const.LB_HEALTH_MONITOR_PING:
        return lb_client.lb_monitor_profile_icmp
    else:
        msg = (_('Cannot create health monitor %(monitor)s with '
                 'type %(type)s') % {'monitor': hm['id'],
                                     'type': hm['type']})
        raise n_exc.InvalidInput(error_message=msg)


def get_tags(plugin, resource_id, resource_type, project_id, project_name):
    return lb_utils.get_tags(plugin, resource_id, resource_type,
                             project_id, project_name)


def build_persistence_profile_tags(pool_tags, listener):
    tags = pool_tags[:]
    # With octavia loadbalancer name might not be among data passed
    # down to the driver
    lb_data = listener.get('loadbalancer')
    if lb_data:
        tags.append({
            'scope': lb_const.LB_LB_NAME,
            'tag': lb_data['name'][:utils.MAX_TAG_LEN]})
    tags.append({
        'scope': lb_const.LB_LB_TYPE,
        'tag': listener['loadbalancer_id']})
    tags.append({
        'scope': lb_const.LB_LISTENER_TYPE,
        'tag': listener['id']})
    return tags


def delete_persistence_profile(nsxpolicy, lb_persistence_profile_path):
    lb_client = nsxpolicy.load_balancer
    pp_client = lb_client.lb_persistence_profile
    persistence_profile_id = p_utils.path_to_id(lb_persistence_profile_path)
    if persistence_profile_id:
        pp_client.delete(persistence_profile_id)


def setup_session_persistence(nsxpolicy, pool, pool_tags, switch_type,
                              listener, vs_data):
    sp = pool.get('session_persistence')
    pers_type = None
    cookie_name = None
    cookie_mode = None
    lb_client = nsxpolicy.load_balancer
    pp_client = None
    if not sp:
        LOG.debug("No session persistence info for pool %s", pool['id'])
    elif sp['type'] == lb_const.LB_SESSION_PERSISTENCE_HTTP_COOKIE:
        pp_client = lb_client.lb_cookie_persistence_profile
        pers_type = nsxlib_lb.PersistenceProfileTypes.COOKIE
        pers_id_suffix = 'cookie'
        cookie_name = sp.get('cookie_name')
        if not cookie_name:
            cookie_name = lb_const.SESSION_PERSISTENCE_DEFAULT_COOKIE_NAME
        cookie_mode = "INSERT"
    elif sp['type'] == lb_const.LB_SESSION_PERSISTENCE_APP_COOKIE:
        pp_client = lb_client.lb_cookie_persistence_profile
        pers_type = nsxlib_lb.PersistenceProfileTypes.COOKIE
        pers_id_suffix = 'cookie'
        # In this case cookie name is mandatory
        cookie_name = sp['cookie_name']
        cookie_mode = "REWRITE"
    else:
        pp_client = lb_client.lb_source_ip_persistence_profile
        pers_type = nsxlib_lb.PersistenceProfileTypes.SOURCE_IP
        pers_id_suffix = 'sourceip'
    if pers_type:
        # There is a profile to create or update
        pp_kwargs = {
            'name': "persistence_%s" % utils.get_name_and_uuid(
                pool['name'] or 'pool', pool['id'], maxlen=235),
            'tags': lb_utils.build_persistence_profile_tags(
                pool_tags, listener)
        }
        if cookie_name:
            pp_kwargs['cookie_name'] = cookie_name
            pp_kwargs['cookie_mode'] = cookie_mode

    profile_path = vs_data.get('lb_persistence_profile_path', '')
    persistence_profile_id = p_utils.path_to_id(profile_path)
    if persistence_profile_id and not switch_type:
        # NOTE: removal of the persistence profile must be executed
        # after the virtual server has been updated
        if pers_type:
            # Update existing profile
            LOG.debug("Updating persistence profile %(profile_id)s for "
                      "listener %(listener_id)s with pool %(pool_id)s",
                      {'profile_id': persistence_profile_id,
                       'listener_id': listener['id'],
                       'pool_id': pool['id']})
            pp_client.update(persistence_profile_id, **pp_kwargs)
            return persistence_profile_id, None
        else:
            # Prepare removal of persistence profile
            return (None, functools.partial(delete_persistence_profile,
                                            nsxpolicy, profile_path))
    elif pers_type:
        # Create persistence profile
        pp_id = "%s_%s" % (pool['id'], pers_id_suffix)
        pp_kwargs['persistence_profile_id'] = pp_id
        pp_client.create_or_overwrite(**pp_kwargs)
        LOG.debug("Created persistence profile %(profile_id)s for "
                  "listener %(listener_id)s with pool %(pool_id)s",
                  {'profile_id': pp_id,
                   'listener_id': listener['id'],
                   'pool_id': pool['id']})
        if switch_type:
            # There is aso a persistence profile to remove!
            return (pp_id, functools.partial(delete_persistence_profile,
                                             nsxpolicy, profile_path))

        return pp_id, None
    return None, None



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from vmware_nsx.db import nsxv_db
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.nsx_v.vshield import edge_firewall_driver
from vmware_nsx.services.fwaas.common import fwaas_callbacks_v2 as \
    com_callbacks
from vmware_nsx.services.fwaas.nsx_tv import edge_fwaas_driver_v2 as tv_driver

LOG = logging.getLogger(__name__)
RULE_NAME_PREFIX = 'Fwaas-'


class NsxvFwaasCallbacksV2(com_callbacks.NsxFwaasCallbacksV2):
    """NSX-V RPC callbacks for Firewall As A Service - V2."""

    def __init__(self, with_rpc):
        super(NsxvFwaasCallbacksV2, self).__init__(with_rpc)
        # update the fwaas driver in case of TV plugin
        self.internal_driver = None
        if self.fwaas_enabled:
            if self.fwaas_driver.driver_name == tv_driver.FWAAS_DRIVER_NAME:
                self.internal_driver = self.fwaas_driver.get_V_driver()
            else:
                self.internal_driver = self.fwaas_driver

    @property
    def plugin_type(self):
        return projectpluginmap.NsxPlugins.NSX_V

    def should_apply_firewall_to_router(self, context, router, router_id):
        """Return True if the FWaaS rules should be added to this router."""
        # in case of a distributed-router:
        # router['id'] is the id of the neutron router (=tlr)
        # and router_id is the plr/tlr (the one that is being updated)

        # First check if there are rules attached to this router
        if not super(NsxvFwaasCallbacksV2,
                     self).should_apply_firewall_to_router(
            context, router['id']):
            return False

        # get all the relevant router info
        # ("router" does not have all the fields)
        ctx_elevated = context.elevated()
        router_data = self.core_plugin.get_router(ctx_elevated, router['id'])
        if not router_data:
            LOG.error("Couldn't read router %s data", router['id'])
            return False

        if router_data.get('distributed'):
            if router_id == router['id']:
                # Do not add firewall rules on the tlr router.
                return False

        # Check if the FWaaS driver supports this router
        if not self.internal_driver.should_apply_firewall_to_router(
            router_data, raise_exception=False):
            return False

        return True

    def get_fwaas_rules_for_router(self, context, router_id, router_db,
                                   edge_id):
        """Return the list of (translated) FWaaS rules for this router."""
        ctx_elevated = context.elevated()
        router_interfaces = self.core_plugin._get_router_interfaces(
            ctx_elevated, router_id)

        fw_rules = []
        # Add firewall rules per port attached to a firewall group
        for port in router_interfaces:
            fwg = self.get_port_fwg(ctx_elevated, port['id'])
            if fwg:
                router_dict = {}
                self.core_plugin._extend_nsx_router_dict(
                    router_dict, router_db)
                if router_dict['distributed']:
                    # The vnic_id is ignored for distributed routers, so
                    # each rule will be applied to all the interfaces.
                    vnic_id = None
                else:
                    # get the interface vnic
                    edge_vnic_bind = nsxv_db.get_edge_vnic_binding(
                        context.session, edge_id, port['network_id'])
                    vnic_id = 'vnic-index-%s' % edge_vnic_bind.vnic_index
                # Add the FWaaS rules for this port
                fw_rules.extend(
                    self.get_port_translated_rules(vnic_id, fwg))

        return fw_rules

    def get_port_translated_rules(self, vnic_id, firewall_group):
        """Return the list of translated rules per port

        Ingress/Egress firewall rules + default ingress/egress drop
        """
        port_rules = []
        logged = False
        # Add the firewall group ingress/egress rules only if the fw is up
        if firewall_group['admin_state_up']:
            port_rules.extend(self.translate_rules(
                firewall_group['ingress_rule_list'],
                replace_dest=vnic_id,
                logged=logged,
                is_ingress=True))
            port_rules.extend(self.translate_rules(
                firewall_group['egress_rule_list'],
                replace_src=vnic_id,
                logged=logged,
                is_ingress=False))

        # Add ingress/egress block rules for this port
        default_ingress = {'name': "Block port ingress",
             'action': edge_firewall_driver.FWAAS_DENY,
             'logged': logged}
        default_egress = {'name': "Block port egress",
             'action': edge_firewall_driver.FWAAS_DENY,
             'logged': logged}
        if vnic_id:
            default_ingress['destination_vnic_groups'] = [vnic_id]
            default_egress['source_vnic_groups'] = [vnic_id]
        port_rules.extend([default_ingress, default_egress])

        return port_rules

    def translate_rules(self, fwaas_rules, replace_dest=None, replace_src=None,
                        logged=False, is_ingress=True):
        translated_rules = []
        for rule in fwaas_rules:
            if not rule['enabled']:
                # skip disabled rules
                continue
            # Make sure the rule has a name, and it starts with the prefix
            # (backend max name length is 30)
            if rule.get('name'):
                rule['name'] = RULE_NAME_PREFIX + rule['name']
            else:
                rule['name'] = RULE_NAME_PREFIX + rule['id']
            rule['name'] = rule['name'][:30]
            if rule.get('id'):
                # update rules ID to prevent DB duplications in
                # NsxvEdgeFirewallRuleBinding
                if is_ingress:
                    rule['id'] = ('ingress-%s' % rule['id'])[:36]
                else:
                    rule['id'] = ('egress-%s' % rule['id'])[:36]
            # source & destination should be lists
            if (rule.get('destination_ip_address') and
                not rule['destination_ip_address'].startswith('0.0.0.0')):
                rule['destination_ip_address'] = [
                    rule['destination_ip_address']]
            else:
                if replace_dest:
                    rule['destination_vnic_groups'] = [replace_dest]
                if 'destination_ip_address' in rule:
                    del rule['destination_ip_address']
            if (rule.get('source_ip_address') and
                not rule['source_ip_address'].startswith('0.0.0.0')):
                rule['source_ip_address'] = [rule['source_ip_address']]
            else:
                if replace_src:
                    rule['source_vnic_groups'] = [replace_src]
                if 'source_ip_address' in rule:
                    del rule['source_ip_address']
            if logged:
                rule['logged'] = True
            translated_rules.append(rule)

        return translated_rules



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.db import db as nsx_db
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_common
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_v3.implementation import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgePoolManagerFromDict(base_mgr.Nsxv3LoadbalancerBaseManager):
    def _get_pool_kwargs(self, name=None, tags=None, algorithm=None,
                         description=None):
        kwargs = {}
        if name:
            kwargs['display_name'] = name
        if tags:
            kwargs['tags'] = tags
        if algorithm:
            kwargs['algorithm'] = algorithm
        if description:
            kwargs['description'] = description
        kwargs['snat_translation'] = {'type': "LbSnatAutoMap"}
        return kwargs

    def _process_vs_update(self, context, pool, switch_type, listener,
                           nsx_pool_id, nsx_vs_id, completor):
        LOG.debug("Processing NSX virtual server update for pool %(pool_id)s. "
                  "Will update VS %(nsx_vs_id)s", {'pool_id': pool['id'],
                                                   'nsx_vs_id': nsx_vs_id})
        vs_client = self.core_plugin.nsxlib.load_balancer.virtual_server
        try:
            # Process pool persistence profile and
            # create/update/delete profile for virtual server
            vs_data = vs_client.get(nsx_vs_id)
            if nsx_pool_id:
                (persistence_profile_id,
                 post_process_func) = lb_utils.setup_session_persistence(
                    self.core_plugin.nsxlib,
                    pool, self._get_pool_tags(context, pool),
                    switch_type, listener, vs_data)
            else:
                post_process_func = functools.partial(
                    lb_utils.delete_persistence_profile,
                    self.core_plugin.nsxlib,
                    vs_data.get('persistence_profile_id'))
                persistence_profile_id = None
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error("Failed to configure session persistence "
                          "profile for pool %(pool_id)s",
                          {'pool_id': pool['id']})
        try:
            # Update persistence profile and pool on virtual server
            vs_client.update(nsx_vs_id, pool_id=nsx_pool_id,
                             persistence_profile_id=persistence_profile_id)
            LOG.debug("Updated NSX virtual server %(vs_id)s with "
                      "pool %(pool_id)s and persistence profile %(prof)s",
                      {'vs_id': nsx_vs_id, 'pool_id': nsx_pool_id,
                       'prof': persistence_profile_id})
            if post_process_func:
                post_process_func()
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to attach pool %s to virtual '
                          'server %s', nsx_pool_id, nsx_vs_id)

    def _get_pool_tags(self, context, pool):
        return lb_utils.get_pool_tags(context, self.core_plugin, pool)

    def create(self, context, pool, completor):
        lb_id = pool['loadbalancer_id']
        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        pool_name = utils.get_name_and_uuid(pool['name'] or 'pool', pool['id'])
        tags = self._get_pool_tags(context, pool)
        description = pool.get('description')
        lb_algorithm = lb_const.LB_POOL_ALGORITHM_MAP.get(pool['lb_algorithm'])
        if pool.get('listeners') and len(pool['listeners']) > 1:
            completor(success=False)
            msg = (_('Failed to create pool: Multiple listeners are not '
                     'supported.'))
            raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

        # NOTE(salv-orlando): Guard against accidental compat breakages
        try:
            listener = pool['listener'] or pool['listeners'][0]
        except IndexError:
            # If listeners is an empty list we hit this exception
            listener = None
        # Perform additional validation for session persistence before
        # creating resources in the backend
        lb_common.validate_session_persistence(pool, listener, completor)
        try:
            kwargs = self._get_pool_kwargs(pool_name, tags, lb_algorithm,
                                           description)
            lb_pool = pool_client.create(**kwargs)
            nsx_db.add_nsx_lbaas_pool_binding(
                context.session, lb_id, pool['id'], lb_pool['id'])
        except nsxlib_exc.ManagerError:
            completor(success=False)
            msg = (_('Failed to create pool on NSX backend: %(pool)s') %
                   {'pool': pool['id']})
            raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

        # The pool object can be created with either --listener or
        # --loadbalancer option. If listener is present, the virtual server
        # will be updated with the pool. Otherwise, just return. The binding
        # will be added later when the pool is associated with layer7 rule.
        # FIXME(salv-orlando): This two-step process can leave a zombie pool on
        # NSX if the VS update operation fails
        if listener:
            listener_id = listener['id']
            binding = nsx_db.get_nsx_lbaas_listener_binding(
                context.session, lb_id, listener_id)
            if binding:
                vs_id = binding['lb_vs_id']
                # Updae the virtual server only if it exists
                if vs_id:
                    self._process_vs_update(context, pool, False, listener,
                                            lb_pool['id'], vs_id, completor)
                    nsx_db.update_nsx_lbaas_pool_binding(
                        context.session, lb_id, pool['id'], vs_id)
            else:
                completor(success=False)
                msg = (_("Couldn't find binding on the listener: %s") %
                       listener['id'])
                raise nsx_exc.NsxPluginException(err_msg=msg)
        completor(success=True)

    def update(self, context, old_pool, new_pool, completor):
        pool_client = self.core_plugin.nsxlib.load_balancer.pool
        pool_name = None
        tags = None
        lb_algorithm = None
        description = None
        if new_pool['name'] != old_pool['name']:
            pool_name = utils.get_name_and_uuid(new_pool['name'] or 'pool',
                                                new_pool['id'])
            tags = self._get_pool_tags(context, new_pool)
        if new_pool['lb_algorithm'] != old_pool['lb_algorithm']:
            lb_algorithm = lb_const.LB_POOL_ALGORITHM_MAP.get(
                new_pool['lb_algorithm'])
        if new_pool.get('description') != old_pool.get('description'):
            description = new_pool['description']
        binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, old_pool['loadbalancer_id'], old_pool['id'])
        if not binding:
            completor(success=False)
            msg = (_('Cannot find pool %(pool)s binding on NSX db '
                     'mapping') % {'pool': old_pool['id']})
            raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)
        if new_pool.get('listeners') and len(new_pool['listeners']) > 1:
            completor(success=False)
            msg = (_('Failed to update pool %s: Multiple listeners are not '
                     'supported.') % new_pool['id'])
            raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)
        # NOTE(salv-orlando): Guard against accidental compat breakages
        try:
            listener = new_pool['listener'] or new_pool['listeners'][0]
        except IndexError:
            # If listeners is an empty list we hit this exception
            listener = None

        # Perform additional validation for session persistence before
        # operating on resources in the backend
        lb_common.validate_session_persistence(new_pool, listener, completor)
        try:
            lb_pool_id = binding['lb_pool_id']
            kwargs = self._get_pool_kwargs(pool_name, tags, lb_algorithm,
                                           description)
            pool_client.update(lb_pool_id, **kwargs)
            # Update virtual servers if it exists and there were changes
            # in session persistence
            if (listener and new_pool['session_persistence'] !=
                old_pool['session_persistence'] and binding['lb_vs_id']):
                switch_type = lb_common.session_persistence_type_changed(
                    new_pool, old_pool)
                self._process_vs_update(context, new_pool, switch_type,
                                        listener, lb_pool_id,
                                        binding['lb_vs_id'], completor)
            completor(success=True)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to update pool %(pool)s with '
                          'error %(error)s',
                          {'pool': old_pool['id'], 'error': e})

    def delete(self, context, pool, completor):
        lb_id = pool['loadbalancer_id']
        pool_client = self.core_plugin.nsxlib.load_balancer.pool

        binding = nsx_db.get_nsx_lbaas_pool_binding(
            context.session, lb_id, pool['id'])
        if binding:
            vs_id = binding.get('lb_vs_id')
            lb_pool_id = binding.get('lb_pool_id')

            if vs_id:
                # NOTE(salv-orlando): Guard against accidental compat breakages
                try:
                    listener = pool['listener'] or pool['listeners'][0]
                except IndexError:
                    # If listeners is an empty list we hit this exception
                    listener = None
                if listener:
                    self._process_vs_update(context, pool, False, listener,
                                            None, vs_id, completor)

            try:
                pool_client.delete(lb_pool_id)
            except nsxlib_exc.ResourceNotFound:
                pass
            except nsxlib_exc.ManagerError:
                completor(success=False)
                msg = (_('Failed to delete lb pool from nsx: %(pool)s') %
                       {'pool': lb_pool_id})
                raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)
            nsx_db.delete_nsx_lbaas_pool_binding(context.session,
                                                 lb_id, pool['id'])

            # Delete the attached health monitor as well
            if pool.get('healthmonitor'):
                hm = pool['healthmonitor']
                monitor_client = self.core_plugin.nsxlib.load_balancer.monitor

                hm_binding = nsx_db.get_nsx_lbaas_monitor_binding(
                    context.session, lb_id, pool['id'], hm['id'])
                if hm_binding:
                    lb_monitor_id = hm_binding['lb_monitor_id']
                    try:
                        monitor_client.delete(lb_monitor_id)
                    except nsxlib_exc.ResourceNotFound:
                        pass
                    except nsxlib_exc.ManagerError as exc:
                        completor(success=False)
                        msg = _('Failed to delete monitor %(monitor)s from '
                                'backend with exception %(exc)s') % {
                                'monitor': hm['id'], 'exc': exc}
                        raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

                    nsx_db.delete_nsx_lbaas_monitor_binding(
                        context.session, lb_id, pool['id'], hm['id'])

        completor(success=True)

    def delete_cascade(self, context, pool, completor):
        self.delete(context, pool, completor)
